@@ -1,26 +1,35 @@
 """
-    flaskext.frozen
-    ~~~~~~~~~~~~~~~
+    flask_frozen
+    ~~~~~~~~~~~~
 
     Frozen-Flask freezes a Flask application into a set of static files.
     The result can be hosted without any server-side software other than a
     traditional web server.
 
 
-    :copyright: (c) 2010 by Simon Sapin.
+    :copyright: (c) 2010-2012 by Simon Sapin.
     :license: BSD, see LICENSE for more details.
+
 """
 
 from __future__ import with_statement
 
+__all__ = ['Freezer', 'walk_directory', 'relative_url_for']
+
+VERSION = '0.10'
+
+
 import os.path
 import mimetypes
 import urlparse
 import urllib
 import warnings
 import collections
+import posixpath
+from fnmatch import fnmatch
 from unicodedata import normalize
 from threading import Lock
+from contextlib import contextmanager
 
 from werkzeug.exceptions import HTTPException
 from flask import (Flask, Blueprint, url_for, request, send_from_directory,
@@ -35,10 +44,22 @@ except ImportError:
     def is_mapping(obj):
         return hasattr(obj, 'keys')
 
+try:
+    from posixpath import relpath as posix_relpath
+except ImportError:
+    # Python 2.5
+    def posix_relpath(path, start):
+        sep = posixpath.sep
+        start_list = [x for x in posixpath.abspath(start).split(sep) if x]
+        path_list = [x for x in posixpath.abspath(path).split(sep) if x]
 
-__all__ = ['Freezer']
+        # Work out how much of the filepath is shared by start and path.
+        i = len(posixpath.commonprefix([start_list, path_list]))
 
-VERSION = '0.9'
+        rel_list = [posixpath.pardir] * (len(start_list)-i) + path_list[i:]
+        if not rel_list:
+            return posixpath.curdir
+        return posixpath.join(*rel_list)
 
 
 class MissingURLGeneratorWarning(Warning):
@@ -86,15 +107,15 @@ class Freezer(object):
         """
         self.app = app
         if app:
-            logger_class = (UrlForLogger if self.log_url_for
-                            else DummyUrlForLogger)
-            self.url_for_logger = logger_class(app)
+            self.url_for_logger = UrlForLogger(app)
             app.config.setdefault('FREEZER_DESTINATION', 'build')
+            app.config.setdefault('FREEZER_DESTINATION_IGNORE', [])
             app.config.setdefault('FREEZER_BASE_URL', 'http://localhost/')
             app.config.setdefault('FREEZER_REMOVE_EXTRA_FILES', True)
             app.config.setdefault('FREEZER_DEFAULT_MIMETYPE',
                                   'application/octet-stream')
             app.config.setdefault('FREEZER_IGNORE_MIMETYPE_WARNINGS', False)
+            app.config.setdefault('FREEZER_RELATIVE_URLS', False)
 
     def register_generator(self, function):
         """Register a function as an URL generator.
@@ -111,7 +132,10 @@ class Freezer(object):
 
     @property
     def root(self):
-        """The build destination."""
+        """
+        Absolute path to the directory Frozen-Flask writes to,
+        ie. resolved value for the ``FREEZER_DESTINATION`` configuration_.
+        """
         # unicode() will raise if the path is not ASCII or already unicode.
         return os.path.join(
             unicode(self.app.root_path),
@@ -123,14 +147,16 @@ class Freezer(object):
         remove_extra = self.app.config['FREEZER_REMOVE_EXTRA_FILES']
         if not os.path.isdir(self.root):
             os.makedirs(self.root)
-        previous_files = set(
-            # See https://github.com/SimonSapin/Frozen-Flask/issues/5
-            normalize('NFC', os.path.join(self.root, *name.split('/')))
-            for name in walk_directory(self.root)
-        )
+        if remove_extra:
+            ignore = self.app.config['FREEZER_DESTINATION_IGNORE']
+            previous_files = set(
+                # See https://github.com/SimonSapin/Frozen-Flask/issues/5
+                normalize('NFC', os.path.join(self.root, *name.split('/')))
+                for name in walk_directory(self.root, ignore=ignore))
         seen_urls = set()
         seen_endpoints = set()
         built_files = set()
+
         for url, endpoint in self._generate_all_urls():
             seen_endpoints.add(endpoint)
             if url in seen_urls:
@@ -139,6 +165,7 @@ class Freezer(object):
             seen_urls.add(url)
             new_filename = self._build_one(url)
             built_files.add(normalize('NFC', new_filename))
+
         self._check_endpoints(seen_endpoints)
         if remove_extra:
             # Remove files from the previous build that are not here anymore.
@@ -217,9 +244,10 @@ class Freezer(object):
         """
         Warn if some of the app's enpoints are not in seen_endpoints.
         """
-        all_endpoints = set(
-            rule.endpoint for rule in self.app.url_map.iter_rules())
-        not_generated_endpoints = all_endpoints - seen_endpoints
+        get_endpoints = set(
+            rule.endpoint for rule in self.app.url_map.iter_rules()
+            if 'GET' in rule.methods)
+        not_generated_endpoints = get_endpoints - seen_endpoints
 
         if self.static_files_urls in self.url_generators:
             # Special case: do not warn when there is no static file
@@ -239,9 +267,11 @@ class Freezer(object):
         client = self.app.test_client()
         base_url = self.app.config['FREEZER_BASE_URL']
 
-        with self.url_for_logger:
-            response = client.get(url, follow_redirects=True,
-                                  base_url=base_url)
+        with conditional_context(self.url_for_logger, self.log_url_for):
+            with conditional_context(patch_url_for(self.app),
+                                     self.app.config['FREEZER_RELATIVE_URLS']):
+                response = client.get(url, follow_redirects=True,
+                                      base_url=base_url)
 
         # The client follows redirects by itself
         # Any other status code is probably an error
@@ -369,24 +399,91 @@ class Freezer(object):
     def no_argument_rules_urls(self):
         """URL generator for URL rules that take no arguments."""
         for rule in self.app.url_map.iter_rules():
-            if not rule.arguments:
+            if not rule.arguments and 'GET' in rule.methods:
                 yield rule.endpoint, {}
 
 
-def walk_directory(root):
+def walk_directory(root, ignore=()):
     """
     Recursively walk the `root` directory and yield slash-separated paths
     relative to the root.
 
-    Used to implement the URL genertor for static files.
+    Used to implement the URL generator for static files.
+
+    :param ignore:
+        A list of :mod:`fnmatch` patterns.
+        As in ``.gitignore`` files,
+        patterns that contains a slash are matched against the whole path,
+        others against individual slash-separated parts.
+
     """
-    for name in os.listdir(root):
-        full_name = os.path.join(root, name)
-        if os.path.isdir(full_name):
-            for filename in walk_directory(full_name):
-                yield name + '/' + filename
-        elif os.path.isfile(full_name):
-            yield name
+    path_ignore = [n.strip('/') for n in ignore if '/' in n]
+    basename_ignore = [n for n in ignore if '/'  not in n]
+
+    def walk(directory, path_so_far):
+        for name in sorted(os.listdir(directory)):
+            if any(fnmatch(name, pattern) for pattern in basename_ignore):
+                continue
+            path = path_so_far + '/' + name if path_so_far else name
+            if any(fnmatch(path, pattern) for pattern in path_ignore):
+                continue
+            full_name = os.path.join(directory, name)
+            if os.path.isdir(full_name):
+                for file_path in walk(full_name, path):
+                    yield file_path
+            elif os.path.isfile(full_name):
+                yield path
+    return walk(root, '')
+
+
+@contextmanager
+def patch_url_for(app):
+    """Patche ``url_for`` in Jinja globals to use :func:`relative_url_for`.
+
+    This is a context manager, to use in a ``with`` statement.
+
+    """
+    previous_url_for = app.jinja_env.globals['url_for']
+    app.jinja_env.globals['url_for'] = relative_url_for
+    try:
+        yield
+    finally:
+        app.jinja_env.globals['url_for'] = previous_url_for
+
+
+def relative_url_for(endpoint, **values):
+    """
+    Like :func:`~flask.url_for`, but returns relative URLs if possible.
+
+    Absolute URLs (with ``_external=True`` or to a different subdomain) are
+    unchanged, but eg. ``/foo/bar`` becomes ``../bar``, depending on the
+    current request context's path. (This, of course, requires a Flask
+    :ref:`request context <flask:request-context>`.)
+
+    URLs that would otherwise end with ``/`` get ``index.html`` appended,
+    as Frozen-Flask does in filenames. Because of this behavior, this function
+    should only with Frozen-Flask, not when running the application in
+    :meth:`app.run() <flask.Flask.run>` or another WSGI sever.
+
+    If the ``FREEZER_RELATIVE_URLS`` `configuration`_ is True, Frozen-Flask
+    will automatically patch the application's Jinja environment so that
+    ``url_for`` in templates is this function.
+
+    """
+    url = url_for(endpoint, **values)
+
+    # absolute URLs in http://... (with subdomains or _external=True)
+    if not url.startswith('/'):
+        return url
+
+    if url.endswith('/'):
+        url += 'index.html'
+
+    request_path = request.path
+    if not request_path.endswith('/'):
+        request_path = posixpath.dirname(request_path)
+
+    return posix_relpath(url, request_path)
 
 
 def unwrap_method(method):
@@ -413,6 +510,16 @@ def method_self(method):
         return method.__self__
 
 
+@contextmanager
+def conditional_context(context, condition):
+    """Wrap a context manager but only enter/exit it if condition is true."""
+    if condition:
+        with context:
+            yield
+    else:
+        yield
+
+
 class UrlForLogger(object):
     """
     Log all calls to url_for() for this app made inside the with block.
@@ -454,23 +561,6 @@ class UrlForLogger(object):
             yield self.logged_calls.popleft()
 
 
-class DummyUrlForLogger(object):
-    """
-    Gives the same API as UrlForLogger, but does not actually log anything.
-    """
-    def __init__(self, app):
-        pass
-
-    def __enter__(self):
-        pass
-
-    def __exit__(self, exc_type, exc_value, traceback):
-        pass
-
-    def iter_calls(self):
-        return iter([])
-
-
 def script_name_middleware(application, script_name):
     """
     Wrap a WSGI application in a middleware that moves ``script_name``
