@@ -1,10 +1,29 @@
 # coding: utf8
+"""
+    flask_frozen.test_app
+    ~~~~~~~~~~~~~~~~~~~~~
+
+    Test application Frozen-Flask
+
+    :copyright: (c) 2010-2012 by Simon Sapin.
+    :license: BSD, see LICENSE for more details.
+
+"""
+
+from __future__ import with_statement
+
+import os.path
+from functools import partial
+
 from flask import Flask, url_for
 from flask_frozen import Freezer
 
 from .admin import admin_blueprint
 
 
+FAVICON = os.path.join(os.path.dirname(__file__), 'static', 'favicon.ico')
+
+
 def create_app(defer_init_app=False, freezer_kwargs=None):
     app = Flask(__name__)
     app.register_blueprint(admin_blueprint, url_prefix='/admin')
@@ -36,22 +55,30 @@ def create_app(defer_init_app=False, freezer_kwargs=None):
         content = 'User-agent: *\nDisallow: /'
         return app.response_class(content, mimetype='text/plain')
 
+    for asset in ("favicon.ico",):
+        url = "/" + asset
+        name = asset.replace(".", "_")
+        app.add_url_rule(url, name, partial(app.send_static_file, filename=asset))
+
     @app.route('/product_<int:product_id>/')
     def product(product_id):
         return 'Product num %i' % product_id
 
+    @app.route('/add/', methods=['POST'])
+    def add_something(product_id):
+        return 'This view should be ignored as it does not accept GET.'
+
     @freezer.register_generator
     def product():
         # endpoint, values
         yield 'product', {'product_id': 0}
+        yield 'page', {'name': 'foo'}
         # Just a `values` dict. The endpoint defaults to the name of the
         # generator function, just like with Flask views
         yield {'product_id': 1}
         # single string: url
         yield '/product_2/'
 
-        yield 'page', {'name': u'I løvë Unicode'}
-
     if defer_init_app:
         freezer.init_app(app)
 
